#!/usr/bin/env python3
"""Test for the ee.filter module."""

import datetime
import json
from typing import Any, Dict

import unittest
import ee
from ee import apitestcase


def make_expression_graph(
    function_invocation_value: Dict[str, Any],
) -> Dict[str, Any]:
  return {
      'result': '0',
      'values': {'0': {'functionInvocationValue': function_invocation_value}},
  }


def max_error_expression(max_error: float) -> Dict[str, Any]:
  return {
      'functionInvocationValue': {
          'functionName': 'ErrorMargin',
          'arguments': {'value': {'constantValue': max_error}},
      }
  }


class FilterTest(apitestcase.ApiTestCase):

  def test_constructors(self):
    """Verifies that constructors understand valid parameters."""
    from_static_method = ee.Filter.gt('foo', 1)
    from_computed_object = ee.Filter(
        ee.ApiFunction.call_('Filter.greaterThan', 'foo', 1))
    self.assertEqual(from_static_method, from_computed_object)

    copy = ee.Filter(from_static_method)
    self.assertEqual(from_static_method, copy)

  def test_metadata(self):
    """Verifies that the metadata_() method works."""
    self.assertEqual(
        ee.ApiFunction.call_('Filter.equals', 'x', 1),
        ee.Filter.metadata_('x', 'equals', 1))
    self.assertEqual(
        ee.Filter.metadata_('x', 'equals', 1), ee.Filter.eq('x', 1))
    self.assertEqual(
        ee.Filter.metadata_('x', 'EQUALS', 1), ee.Filter.eq('x', 1))
    self.assertEqual(
        ee.Filter.metadata_('x', 'not_equals', 1), ee.Filter.neq('x', 1))
    self.assertEqual(
        ee.Filter.metadata_('x', 'less_than', 1), ee.Filter.lt('x', 1))
    self.assertEqual(
        ee.Filter.metadata_('x', 'not_greater_than', 1), ee.Filter.lte('x', 1))
    self.assertEqual(
        ee.Filter.metadata_('x', 'greater_than', 1), ee.Filter.gt('x', 1))
    self.assertEqual(
        ee.Filter.metadata_('x', 'not_less_than', 1), ee.Filter.gte('x', 1))

  def test_logical_combinations(self):
    """Verifies that the and() and or() methods work."""
    f1 = ee.Filter.eq('x', 1)
    f2 = ee.Filter.eq('x', 2)

    or_filter = ee.Filter.Or(f1, f2)
    self.assertEqual(ee.ApiFunction.call_('Filter.or', (f1, f2)), or_filter)

    and_filter = ee.Filter.And(f1, f2)
    self.assertEqual(ee.ApiFunction.call_('Filter.and', (f1, f2)), and_filter)

    self.assertEqual(
        ee.ApiFunction.call_('Filter.or', (or_filter, and_filter)),
        ee.Filter.Or(or_filter, and_filter))

  def test_date(self):
    """Verifies that date filters work."""
    d1 = datetime.datetime.strptime('1/1/2000', '%m/%d/%Y')
    d2 = datetime.datetime.strptime('1/1/2001', '%m/%d/%Y')
    instant_range = ee.ApiFunction.call_('DateRange', d1, None)
    long_range = ee.ApiFunction.call_('DateRange', d1, d2)

    instant_filter = ee.Filter.date(d1)
    self.assertEqual(
        ee.ApiFunction.lookup('Filter.dateRangeContains'), instant_filter.func)
    self.assertEqual({
        'leftValue': instant_range,
        'rightField': ee.String('system:time_start')
    }, instant_filter.args)

    long_filter = ee.Filter.date(d1, d2)
    self.assertEqual(
        ee.ApiFunction.lookup('Filter.dateRangeContains'), long_filter.func)
    self.assertEqual({
        'leftValue': long_range,
        'rightField': ee.String('system:time_start')
    }, long_filter.args)

  def test_bounds(self):
    """Verifies that geometry intersection filters work."""
    polygon = ee.Geometry.Polygon(1, 2, 3, 4, 5, 6)
    self.assertEqual(
        ee.ApiFunction.call_('Filter.intersects', '.all',
                             ee.ApiFunction.call_('Feature', polygon)),
        ee.Filter.geometry(polygon))

    # Collection-to-geometry promotion.
    collection = ee.FeatureCollection('foo')
    feature = ee.ApiFunction.call_(
        'Feature', ee.ApiFunction.call_('Collection.geometry', collection))
    self.assertEqual(
        ee.ApiFunction.call_('Filter.intersects', '.all', feature),
        ee.Filter.geometry(collection))

    # Check the bounds() alias.
    self.assertEqual(
        ee.ApiFunction.call_('Filter.intersects', '.all',
                             ee.ApiFunction.call_('Feature', polygon)),
        ee.Filter.bounds(polygon))

  def test_in_list(self):
    """Verifies that list membership filters work."""
    self.assertEqual(
        ee.Filter.listContains(None, None, 'foo', [1, 2]),  # pytype: disable=attribute-error
        ee.Filter.inList('foo', [1, 2]))

  def test_internals(self):
    """Test eq(), ne() and hash()."""
    a = ee.Filter.eq('x', 1)
    b = ee.Filter.eq('x', 2)
    c = ee.Filter.eq('x', 1)

    self.assertEqual(a, a)
    self.assertNotEqual(a, b)
    self.assertEqual(a, c)
    self.assertNotEqual(b, c)
    self.assertNotEqual(hash(a), hash(b))

  def test_init_opt_params(self):
    result = ee.Filter(opt_filter=[ee.Filter.gt('prop', 1)]).serialize()
    self.assertIn('"functionName": "Filter.greaterThan"', result)

  def test_date_opt_params(self):
    result = ee.Filter.date(
        start='1996-01-01T00:00', opt_end='2023-01-01T00:00'
    ).serialize()
    self.assertIn('"end": {"constantValue": "2023-01-01T00:00"}', result)

  def test_in_list_opt_params(self):
    result = ee.Filter.inList(
        opt_leftField='lf',
        opt_rightValue='rv',
        opt_rightField='rf',
        opt_leftValue='lv',
    ).serialize()
    self.assertIn('"leftField": {"constantValue": "rf"}', result)
    self.assertIn('"leftValue": {"constantValue": "rv"}', result)
    self.assertIn('"rightField": {"constantValue": "lf"}', result)
    self.assertIn('"rightValue": {"constantValue": "lv"}', result)

  def test_bounds_opt_params(self):
    result = ee.Filter.bounds(
        geometry=ee.Geometry.Point(0, 0), opt_errorMargin=12345
    ).serialize()
    self.assertIn('"value": {"constantValue": 12345}', result)

  def test_and(self):
    expect = make_expression_graph({
        'arguments': {'filters': {'constantValue': []}},
        'functionName': 'Filter.and',
    })
    expression = ee.Filter.And([])
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    # TODO: ee.Filter.And(filters=[]) does not currently work.

  def test_area(self):
    min_val = 1
    max_val = 2
    max_error = 3
    geometry_selector = 'a'
    expect = make_expression_graph({
        'arguments': {
            'min': {'constantValue': min_val},
            'max': {'constantValue': max_val},
            'maxError': max_error_expression(max_error),
            'geometrySelector': {'constantValue': geometry_selector},
        },
        'functionName': 'Filter.area',
    })
    expression = ee.Filter.area(min_val, max_val, max_error, geometry_selector)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.area(
        min=min_val,
        max=max_val,
        maxError=max_error,
        geometrySelector=geometry_selector,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_calendar_range(self):
    start = 1
    end = 2
    field = 'year'
    expect = make_expression_graph({
        'arguments': {
            'start': {'constantValue': start},
            'end': {'constantValue': end},
            'field': {'constantValue': field},
        },
        'functionName': 'Filter.calendarRange',
    })
    expression = ee.Filter.calendarRange(start, end, field)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.calendarRange(start=start, end=end, field=field)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_contains(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    max_error = 3
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
            'maxError': max_error_expression(max_error),
        },
        'functionName': 'Filter.contains',
    })
    expression = ee.Filter.contains(
        left_field, right_value, right_field, left_value, max_error
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.contains(
        leftField=left_field,
        rightValue=right_value,
        rightField=right_field,
        leftValue=left_value,
        maxError=max_error,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_date_range_contains(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.dateRangeContains',
    })
    expression = ee.Filter.dateRangeContains(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.dateRangeContains(
        leftField=left_field,
        rightValue=right_value,
        rightField=right_field,
        leftValue=left_value,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_day_of_year(self):
    start = 1
    end = 2
    expect = make_expression_graph({
        'arguments': {
            'start': {'constantValue': start},
            'end': {'constantValue': end},
        },
        'functionName': 'Filter.dayOfYear',
    })
    expression = ee.Filter.dayOfYear(start, end)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.dayOfYear(start=start, end=end)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_disjoint(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    max_error = 3
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
            'maxError': max_error_expression(max_error),
        },
        'functionName': 'Filter.disjoint',
    })
    expression = ee.Filter.disjoint(
        left_field, right_value, right_field, left_value, max_error
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.disjoint(
        leftField=left_field,
        rightValue=right_value,
        rightField=right_field,
        leftValue=left_value,
        maxError=max_error,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_eq(self):
    # ee.Filter.eq uses Equals and masks the actual eq call.
    name = 'a'
    value = 1
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': name},
            'rightValue': {'constantValue': value},
        },
        'functionName': 'Filter.equals',
    })
    expression = ee.Filter.eq(name, value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.eq(name=name, value=value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_equals(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.equals',
    })
    expression = ee.Filter.equals(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.equals(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_expression(self):
    expression_str = 'an expression'
    expect = make_expression_graph({
        'arguments': {
            'expression': {'constantValue': expression_str},
        },
        'functionName': 'Filter.expression',
    })
    expression = ee.Filter.expression(expression_str)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.expression(expression=expression_str)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_greater_than(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.greaterThan',
    })
    expression = ee.Filter.greaterThan(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.greaterThan(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_greater_than_or_equals(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.greaterThanOrEquals',
    })
    expression = ee.Filter.greaterThanOrEquals(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.greaterThanOrEquals(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_gt(self):
    # ee.Filter.gt uses greaterThan and masks the actual gt call.
    name = 'a'
    value = 1
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': name},
            'rightValue': {'constantValue': value},
        },
        'functionName': 'Filter.greaterThan',
    })
    expression = ee.Filter.gt(name, value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.gt(name=name, value=value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_gte(self):
    # gte is implemented as lt().not().
    name = 'a'
    value = 1
    expect = make_expression_graph({
        'functionName': 'Filter.not',
        'arguments': {
            'filter': {
                'functionInvocationValue': {
                    'functionName': 'Filter.lessThan',
                    'arguments': {
                        'leftField': {'constantValue': name},
                        'rightValue': {'constantValue': value},
                    },
                }
            }
        },
    })
    expression = ee.Filter.gte(name, value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.gte(name=name, value=value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  # TODO:  test_has_type

  def test_in_list_serialize(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    # Warning: inList swaps around args.
    expect = make_expression_graph({
        'arguments': {
            'rightField': {'constantValue': left_field},
            'leftValue': {'constantValue': right_value},
            'leftField': {'constantValue': right_field},
            'rightValue': {'constantValue': left_value},
        },
        # Note this is not 'inList'.
        'functionName': 'Filter.listContains',
    })
    expression = ee.Filter.inList(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.inList(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_intersects(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    max_error = 3
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
            'maxError': max_error_expression(max_error),
        },
        'functionName': 'Filter.intersects',
    })
    expression = ee.Filter.intersects(
        left_field, right_value, right_field, left_value, max_error
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.intersects(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
        maxError=max_error,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_is_contained(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    max_error = 3
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
            'maxError': max_error_expression(max_error),
        },
        'functionName': 'Filter.isContained',
    })
    expression = ee.Filter.isContained(
        left_field, right_value, right_field, left_value, max_error
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.isContained(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
        maxError=max_error,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_less_than(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.lessThan',
    })
    expression = ee.Filter.lessThan(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.lessThan(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_less_than_or_equals(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.lessThanOrEquals',
    })
    expression = ee.Filter.lessThanOrEquals(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.lessThanOrEquals(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_list_contains(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.listContains',
    })
    expression = ee.Filter.listContains(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.listContains(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_lt(self):
    # Note: not Filter.lt.
    name = 'a'
    value = 1
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': name},
            'rightValue': {'constantValue': value},
        },
        'functionName': 'Filter.lessThan',
    })
    expression = ee.Filter.lt(name, value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.lt(name=name, value=value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_lte(self):
    # Note: not Filter.lte.
    name = 'a'
    value = 1
    expect = make_expression_graph({
        'functionName': 'Filter.not',
        'arguments': {
            'filter': {
                'functionInvocationValue': {
                    'functionName': 'Filter.greaterThan',
                    'arguments': {
                        'leftField': {'constantValue': name},
                        'rightValue': {'constantValue': value},
                    },
                }
            }
        },
    })
    expression = ee.Filter.lte(name, value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.lte(name=name, value=value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_max_difference(self):
    difference = 1
    left_field = 'a'
    right_value = 2
    right_field = 'b'
    left_value = 3
    expect = make_expression_graph({
        'arguments': {
            'difference': {'constantValue': difference},
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.maxDifference',
    })
    expression = ee.Filter.maxDifference(
        difference, left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.maxDifference(
        difference=difference,
        leftField=left_field,
        rightValue=right_value,
        rightField=right_field,
        leftValue=left_value,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_neq(self):
    # Note: Filter.not paired with a Filter.equals.
    name = 'a'
    value = 1
    expect = make_expression_graph({
        'functionName': 'Filter.not',
        'arguments': {
            'filter': {
                'functionInvocationValue': {
                    'functionName': 'Filter.equals',
                    'arguments': {
                        'leftField': {'constantValue': 'a'},
                        'rightValue': {'constantValue': 1},
                    },
                }
            }
        },
    })
    expression = ee.Filter.neq(name, value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.neq(name=name, value=value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_not(self):
    name = 'air_temp'
    value = 2
    gt_filter = ee.Filter.gt(name, value)
    expect = make_expression_graph({
        'functionName': 'Filter.not',
        'arguments': {
            'filter': {
                'functionInvocationValue': {
                    'functionName': 'Filter.greaterThan',
                    'arguments': {
                        'leftField': {'constantValue': name},
                        'rightValue': {'constantValue': value},
                    },
                }
            }
        },
    })
    expression = ee.Filter.Not(gt_filter)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = gt_filter.Not()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_not_equals(self):
    left_field = 'a'
    right_value = 1
    right_field = 'b'
    left_value = 2
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.notEquals',
    })
    expression = ee.Filter.notEquals(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.notEquals(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_not_null(self):
    properties = ['a', 'b']
    expect = make_expression_graph({
        'arguments': {'properties': {'constantValue': properties}},
        'functionName': 'Filter.notNull',
    })
    expression = ee.Filter.notNull(properties)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.notNull(properties=properties)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_or_empty(self):
    expect = make_expression_graph({
        'arguments': {
            'filters': {'constantValue': []},
        },
        'functionName': 'Filter.or',
    })
    expression = ee.Filter.Or()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_or(self):
    name = 'air_temp'
    value = 2
    gt_filter = ee.Filter.gt(name, value)
    expect = make_expression_graph({
        'functionName': 'Filter.or',
        'arguments': {
            'filters': {
                'arrayValue': {
                    'values': [{
                        'functionInvocationValue': {
                            'functionName': 'Filter.greaterThan',
                            'arguments': {
                                'leftField': {'constantValue': name},
                                'rightValue': {'constantValue': value},
                            },
                        }
                    }]
                }
            }
        },
    })

    expression = ee.Filter.Or(gt_filter)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_range_contains(self):
    field = 'a'
    min_value = 1
    max_value = 2
    expect = make_expression_graph({
        'arguments': {
            'field': {'constantValue': field},
            'minValue': {'constantValue': min_value},
            'maxValue': {'constantValue': max_value},
        },
        'functionName': 'Filter.rangeContains',
    })
    expression = ee.Filter.rangeContains(field, min_value, max_value)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.rangeContains(
        field=field, minValue=min_value, maxValue=max_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_string_contains(self):
    left_field = 'a'
    right_value = 'b'
    right_field = 'c'
    left_value = 'd'
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.stringContains',
    })
    expression = ee.Filter.stringContains(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.stringContains(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_string_ends_with(self):
    left_field = 'a'
    right_value = 'b'
    right_field = 'c'
    left_value = 'd'
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.stringEndsWith',
    })
    expression = ee.Filter.stringEndsWith(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.stringEndsWith(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_string_starts_with(self):
    left_field = 'a'
    right_value = 'b'
    right_field = 'c'
    left_value = 'd'
    expect = make_expression_graph({
        'arguments': {
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
        },
        'functionName': 'Filter.stringStartsWith',
    })
    expression = ee.Filter.stringStartsWith(
        left_field, right_value, right_field, left_value
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.stringStartsWith(
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_within_distance(self):
    distance = 1
    left_field = 'a'
    right_value = 2
    right_field = 'b'
    left_value = 3
    max_error = 4
    expect = make_expression_graph({
        'arguments': {
            'distance': {'constantValue': distance},
            'leftField': {'constantValue': left_field},
            'rightValue': {'constantValue': right_value},
            'rightField': {'constantValue': right_field},
            'leftValue': {'constantValue': left_value},
            'maxError': max_error_expression(max_error),
        },
        'functionName': 'Filter.withinDistance',
    })
    expression = ee.Filter.withinDistance(
        distance, left_field, right_value, right_field, left_value, max_error
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Filter.withinDistance(
        distance=distance,
        leftValue=left_value,
        rightValue=right_value,
        leftField=left_field,
        rightField=right_field,
        maxError=max_error,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)


if __name__ == '__main__':
  unittest.main()



"""A wrapper for Reducers."""
from __future__ import annotations

from typing import Any, List, Optional, Tuple, Union

from ee import apifunction
from ee import computedobject
from ee import deprecation
from ee import ee_list
from ee import ee_number
from ee import ee_string
from ee import image

_EeBoolType = Union[Any, computedobject.ComputedObject]
_ImageType = Union[Any, 'image.Image', computedobject.ComputedObject]
_IntegerType = Union[int, 'ee_number.Number', computedobject.ComputedObject]
_ListType = Union[List[Any], Tuple[Any, Any], computedobject.ComputedObject]
_NumberType = Union[float, 'ee_number.Number', computedobject.ComputedObject]
_ReducerType = Union['Reducer', computedobject.ComputedObject]
_StringType = Union[str, 'ee_string.String', computedobject.ComputedObject]


class Reducer(computedobject.ComputedObject):
  """An object to represent an Earth Engine Reducer.

  Example:
    fc = ee.FeatureCollection([
        ee.Feature(None, {'label': 1}),
        ee.Feature(None, {'label': 2}),
    ])
    reducer = ee.Reducer(ee.Reducer.toList())
    result = fc.reduceColumns(reducer, ['label']).get('list')
    print(result.getInfo())
  """

  _initialized: bool = False

  def __init__(
      self,
      reducer: computedobject.ComputedObject,
  ):
    """Creates a Reducer wrapper.

    Args:
      reducer: A reducer to cast.
    """
    self.initialize()

    if isinstance(reducer, computedobject.ComputedObject):
      # There is no server-side constructor for ee.Reducer. Pass the object
      # as-is to the server in case it is intended to be a Reducer cast.
      super().__init__(reducer.func, reducer.args, reducer.varName)
      return

    raise TypeError(
        f'Reducer can only be used as a cast to Reducer. Found {type(reducer)}.'
    )

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  @staticmethod
  def name() -> str:
    return 'Reducer'

  @staticmethod
  def allNonZero() -> Reducer:
    """Returns a Reducer that returns 1 if all inputs are not 0.

    Otherwise, it returns 0.

    Where applicable, the output name is "all".
    """

    return apifunction.ApiFunction.call_('Reducer.allNonZero')

  @staticmethod
  @deprecation.Deprecated('Use Reducer.allNonZero().')
  def And() -> Reducer:
    """Returns a Reducer that returns 1 if all inputs are not 0.

    Otherwise, it returns 0.

    Where applicable, the output name is "all".
    """

    return apifunction.ApiFunction.call_('Reducer.and')

  @staticmethod
  def anyNonZero() -> Reducer:
    """Returns a Reducer that returns 1 if any of its inputs are not 0.

    Otherwise, it returns 0.

    Where applicable, the output name is "any".
    """

    return apifunction.ApiFunction.call_('Reducer.anyNonZero')

  @staticmethod
  def autoHistogram(
      # pylint: disable=invalid-name
      maxBuckets: Optional[_IntegerType] = None,
      minBucketWidth: Optional[_NumberType] = None,
      maxRaw: Optional[_IntegerType] = None,
      # pylint: enable=invalid-name
      cumulative: Optional[_EeBoolType] = None,
  ) -> Reducer:
    """Returns an ee.Reducer that computes a histogram of the inputs.

    The output is a Nx2 array of the lower bucket bounds and the counts (or
    cumulative counts) of each bucket and is suitable for use per-pixel.

    Args:
      maxBuckets: The maximum number of buckets to use when building a
        histogram; will be rounded up to a power of 2.
      minBucketWidth: The minimum histogram bucket width, or null to allow any
        power of 2.
      maxRaw: The number of values to accumulate before building the initial
        histogram.
      cumulative: When true, generates a cumulative histogram.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.autoHistogram',
        maxBuckets,
        minBucketWidth,
        maxRaw,
        cumulative,
    )

  @staticmethod
  def bitwiseAnd() -> Reducer:
    """Returns a Reducer that computes the bitwise-and summation of its inputs.

    Where applicable, the output name is "bitwiseAnd".
    """

    return apifunction.ApiFunction.call_('Reducer.bitwiseAnd')

  @staticmethod
  def bitwiseOr() -> Reducer:
    """Returns a Reducer that computes the bitwise-or summation of its inputs.

    Where applicable, the output name is "bitwiseOr".
    """

    return apifunction.ApiFunction.call_('Reducer.bitwiseOr')

  @staticmethod
  def centeredCovariance() -> Reducer:
    """Returns a centered covariance reducer.

    Creates a reducer that reduces some number of 1-D arrays of the same length
    N to a covariance matrix of shape NxN.

    WARNING: this reducer requires that the data has been mean centered.
    """

    return apifunction.ApiFunction.call_('Reducer.centeredCovariance')

  @staticmethod
  def circularMean() -> Reducer:
    """Returns a circular mean reducer.

    Returns a Reducer that computes the (weighted) circular mean of its inputs,
    which are expected to be in radians.

    Output will be in the range (-π to π). Where applicable, the output name is
    "circularMean".
    """

    return apifunction.ApiFunction.call_('Reducer.circularMean')

  @staticmethod
  def circularStddev() -> Reducer:
    """Returns a circular standard deviation reducer.

    Returns a reducer that computes the (weighted) circular standard deviation
    of its inputs, which are expected to be in radians, using the formula
    sqrt(-2 * ln(R)).

    Where applicable, the output name is "circularStdDev".
    """

    return apifunction.ApiFunction.call_('Reducer.circularStddev')

  @staticmethod
  def circularVariance() -> Reducer:
    """Returns a circular variance Reducer.

    Returns a Reducer that computes the (weighted) circular variance of its
    inputs, which are expected to be in radians.

    Where applicable, the output name is "circularVariance".
    """

    return apifunction.ApiFunction.call_('Reducer.circularVariance')

  def combine(
      self,
      reducer2: _ReducerType,
      # pylint: disable=invalid-name
      outputPrefix: Optional[_StringType] = None,
      sharedInputs: Optional[_EeBoolType] = None,
      # pylint: enable=invalid-name
  ) -> Reducer:
    """Returns a Reducer that runs two reducers in parallel.

    The combined reducer's outputs will be those of reducer1 followed by those
    of reducer2, where the output names of reducer2 are prefixed with the given
    string.

    If sharedInputs is true, the reducers must have the same number of inputs,
    and the combined reducer's will match them; if it is false, the inputs of
    the combined reducer will be those of reducer1 followed by those of
    reducer2.

    Args:
      reducer2: The second reducer.
      outputPrefix: Prefix for reducer2's output names.
      sharedInputs: Whether the reducers share inputs.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.combine',
        self,
        reducer2,
        outputPrefix,
        sharedInputs,
    )

  @staticmethod
  def count() -> Reducer:
    """Returns a Reducer that computes the number of non-null inputs.

    Where applicable, the output name is "count".
    """

    return apifunction.ApiFunction.call_('Reducer.count')

  @staticmethod
  def countDistinct() -> Reducer:
    """Returns a Reducer that computes the number of distinct inputs.

    Where applicable, the output name is "count".
    """

    return apifunction.ApiFunction.call_('Reducer.countDistinct')

  @staticmethod
  def countDistinctNonNull() -> Reducer:
    """Returns a Reducer that computes the number of distinct inputs, ignoring nulls.

    Where applicable, the output name is "count".
    """

    return apifunction.ApiFunction.call_('Reducer.countDistinctNonNull')

  @staticmethod
  def countEvery() -> Reducer:
    """Returns a Reducer that computes the number of inputs.

    Where applicable, the output name is "count".
    """

    return apifunction.ApiFunction.call_('Reducer.countEvery')

  @staticmethod
  def countRuns() -> Reducer:
    """Returns a Reducer that computes the number of runs of distinct, non-null inputs.

    Where applicable, the output name is "count".
    """

    return apifunction.ApiFunction.call_('Reducer.countRuns')

  @staticmethod
  def covariance() -> Reducer:
    """Returns a covariance reducer.

    Creates a reducer that reduces some number of 1-D arrays of the same length
    N to a covariance matrix of shape NxN.

    This reducer uses the one-pass covariance formula from Sandia National
    Laboratories Technical Report SAND2008-6212, which can lose accuracy if the
    values span a large range.
    """

    return apifunction.ApiFunction.call_('Reducer.covariance')

  def disaggregate(self, axis: Optional[_IntegerType] = None) -> Reducer:
    """Returns a Reducer that separates aggregate inputs.

    Separates aggregate inputs (Arrays, Lists, or Dictionaries) into individual
    items that are then each passed to the specified reducer. When used on
    dictionaries, the dictionary keys are ignored. Non-aggregated inputs (e.g.,
    numbers or strings) are passed to the underlying reducer directly.

    Args:
      axis: If specified, indicates an array axis along which to disaggregate.
        If not specified, arrays are completely disaggregated. Ignored for
        non-array types.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.disaggregate', self, axis
    )

  @staticmethod
  def first() -> Reducer:
    """Returns a Reducer that returns the first of its inputs.

    Where applicable, the output name is "first".
    """

    return apifunction.ApiFunction.call_('Reducer.first')

  @staticmethod  # pylint: disable=staticmethod-use
  def firstNonNull() -> Reducer:
    """Returns a Reducer that returns the first of its non-null inputs.

    Where applicable, the output name is "first".
    """

    return apifunction.ApiFunction.call_('Reducer.firstNonNull')

  @staticmethod
  def fixed2DHistogram(
      # pylint: disable=invalid-name
      xMin: _NumberType,
      xMax: _NumberType,
      xSteps: _IntegerType,
      yMin: _NumberType,
      yMax: _NumberType,
      ySteps: _IntegerType,
      # pylint: enable=invalid-name
  ) -> Reducer:
    """Returns a fixed 2D histogram reducer.

    Creates a reducer that will compute a 2D histogram of the inputs using a
    fixed number of fixed width bins. Values outside of the [min, max) range on
    either axis are ignored. The output is a 2D array of counts, and 2 1-D
    arrays of bucket lower edges for the xAxis and the yAxis. This reducer is
    suitable for use per-pixel, however it is always unweighted. The maximum
    count for any bucket is 2^31 - 1.

    Args:
      xMin: The lower (inclusive) bound of the first bucket on the X axis.
      xMax: The upper (exclusive) bound of the last bucket on the X axis.
      xSteps: The number of buckets to use on the X axis.
      yMin: The lower (inclusive) bound of the first bucket on the Y axis.
      yMax: The upper (exclusive) bound of the last bucket on the Y axis.
      ySteps: The number of buckets to use on the Y axis.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.fixed2DHistogram', xMin, xMax, xSteps, yMin, yMax, ySteps
    )

  @staticmethod
  def fixedHistogram(
      min: _NumberType,  # pylint: disable=redefined-builtin
      max: _NumberType,  # pylint: disable=redefined-builtin
      steps: _IntegerType,
      cumulative: Optional[_EeBoolType] = None,
  ) -> Reducer:
    """Returns a fixed histogram reducer.

    Creates a reducer that will compute a histogram of the inputs using a fixed
    number of fixed-width bins. Values outside of the [min, max) range are
    ignored. The output is a Nx2 array of bucket lower edges and counts (or
    cumulative counts) and is suitable for use per-pixel.

    Args:
      min: The lower (inclusive) bound of the first bucket.
      max: The upper (exclusive) bound of the last bucket.
      steps: The number of buckets to use.
      cumulative: When true, generates a cumulative histogram.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.fixedHistogram', min, max, steps, cumulative
    )

  def forEach(
      self,
      outputNames: _ListType,  # pylint: disable=invalid-name
  ) -> Reducer:
    """Returns a reducer that runs a reducer for each output name in the list.

    Creates a Reducer by combining a copy of the given reducer for each output
    name in the given list. If the reducer has a single output, the output names
    are used as-is; otherwise they are prefixed to the original output names.

    Args:
      outputNames: Which output names to use.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.forEach', self, outputNames
    )

  def forEachBand(
      self,
      image: _ImageType,  # pylint: disable=redefined-outer-name
  ) -> Reducer:
    """Returns a reducer that runs a reducer for each band in the image.

    Creates a Reducer by combining a copy of the given reducer for each band in
    the given image, using the band names as output names.

    Args:
      image: The image to use.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.forEachBand', self, image
    )

  def forEachElement(self) -> Reducer:
    """Returns a reducer that runs a reducer for each element in the input.

    Separately reduces each position in array inputs of equal shape, producing
    an array output of the same shape.

    For example, with the 'sum' reducer applied to 5 arrays with shape 2x2, the
    output will be a 2x2 array, where each position is the sum of the 5 values
    at that position.
    """

    return apifunction.ApiFunction.call_(self.name() + '.forEachElement', self)

  @staticmethod
  def frequencyHistogram() -> Reducer:
    """Returns a Reducer that returns a weighted frequency table of its inputs.

    Where applicable, the output name is "histogram".
    """

    return apifunction.ApiFunction.call_('Reducer.frequencyHistogram')

  @staticmethod
  def geometricMedian(
      numX: _IntegerType,  # pylint: disable=invalid-name
      eta: Optional[_NumberType] = None,
      # pylint: disable-next=invalid-name
      initialStepSize: Optional[_NumberType] = None,
  ) -> Reducer:
    """Returns a reducer that computes the geometric median across the inputs.

    Args:
      numX: The number of input dimensions.
      eta: The minimum improvement in the solution used as a stopping criteria
        for the solver.
      initialStepSize: The initial step size used in the solver.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.geometricMedian', numX, eta, initialStepSize
    )

  def getOutputs(self) -> ee_list.List:
    """Returns a list of the output names of the given reducer."""

    return apifunction.ApiFunction.call_(self.name() + '.getOutputs', self)

  def group(
      self,
      groupField: Optional[_IntegerType] = None,
      groupName: Optional[_StringType] = None,
  ) -> Reducer:
    """Returns a reducer groups reducer records by the value of a given input.

    Reduces each group with the given reducer.

    Args:
      groupField: The field that contains record groups.
      groupName: The dictionary key that contains the group. Defaults to
        'group'.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.group', self, groupField, groupName
    )

  @staticmethod
  def histogram(
      # pylint: disable=invalid-name
      maxBuckets: Optional[_IntegerType] = None,
      minBucketWidth: Optional[_NumberType] = None,
      maxRaw: Optional[_IntegerType] = None,
      # pylint: enable=invalid-name
  ) -> Reducer:
    """Returns a reducer that will compute a histogram of the inputs.

    Args:
      maxBuckets: The maximum number of buckets to use when building a
        histogram; will be rounded up to a power of 2.
      minBucketWidth: The minimum histogram bucket width, or null to allow any
        power of 2.
      maxRaw: The number of values to accumulate before building the initial
        histogram.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.histogram', maxBuckets, minBucketWidth, maxRaw
    )

  @staticmethod
  def intervalMean(
      # pylint: disable=invalid-name
      minPercentile: _NumberType,
      maxPercentile: _NumberType,
      maxBuckets: Optional[_IntegerType] = None,
      minBucketWidth: Optional[_NumberType] = None,
      maxRaw: Optional[_IntegerType] = None,
      # pylint: enable=invalid-name
  ) -> Reducer:
    """Returns an interval mean reducer.

    Creates a Reducer to compute the mean of all inputs in the specified
    percentile range. For small numbers of inputs (up to maxRaw) the mean will
    be computed directly; for larger numbers of inputs the mean will be derived
    from a histogram.

    Args:
      minPercentile: The lower bound of the percentile range.
      maxPercentile: The upper bound of the percentile range.
      maxBuckets: The maximum number of buckets to use when building a
        histogram; will be rounded up to a power of 2.
      minBucketWidth: The minimum histogram bucket width, or null to allow any
        power of 2.
      maxRaw: The number of values to accumulate before building the initial
        histogram.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.intervalMean',
        minPercentile,
        maxPercentile,
        maxBuckets,
        minBucketWidth,
        maxRaw,
    )

  @staticmethod
  def kendallsCorrelation(
      numInputs: Optional[_IntegerType] = None,  # pylint: disable=invalid-name
  ) -> Reducer:
    """Returns a reducer that computes the Kendall's Tau-b rank correlation.

    A positive tau value indicates an increasing trend; negative value indicates
    a decreasing trend. See
    https://commons.apache.org/proper/commons-math/javadocs/api-3.6.1/org/apache/commons/math3/stat/correlation/KendallsCorrelation.html
    for details.

    Args:
      numInputs: The number of inputs to expect (1 or 2). If 1 is specified,
        automatically generates sequence numbers for the x value (meaning there
        can be no ties).
    """

    return apifunction.ApiFunction.call_(
        'Reducer.kendallsCorrelation', numInputs
    )

  @staticmethod
  def kurtosis() -> Reducer:
    """Returns a Reducer that Computes the kurtosis of its inputs.

    Where applicable, the output name is "kurtosis".
    """

    return apifunction.ApiFunction.call_('Reducer.kurtosis')

  @staticmethod
  def last() -> Reducer:
    """Returns a Reducer that returns the last of its inputs.

    Where applicable, the output name is "last".
    """

    return apifunction.ApiFunction.call_('Reducer.last')

  @staticmethod
  def lastNonNull() -> Reducer:
    """Returns a Reducer that returns the last of its non-null inputs.

    Where applicable, the output name is "last".
    """

    return apifunction.ApiFunction.call_('Reducer.lastNonNull')

  @staticmethod
  def linearFit() -> Reducer:
    """Returns a linear fit reducer.

    Returns a Reducer that computes the slope and offset for a (weighted) linear
    regression of 2 inputs.

    The inputs are expected to be x data followed by y data. Where applicable,
    the outputs are named: "scale", "offset".
    """

    return apifunction.ApiFunction.call_('Reducer.linearFit')

  @staticmethod
  def linearRegression(
      numX: _IntegerType,  # pylint: disable=invalid-name
      numY: Optional[_IntegerType] = None,  # pylint: disable=invalid-name
  ) -> Reducer:
    """Returns a linear regression reducer.

    Creates a reducer that computes a linear least squares regression with numX
    independent variables and numY dependent variables.

    Each input tuple will have values for the independent variables followed by
    the dependent variables.

    The first output is a coefficients array with dimensions (numX, numY); each
    column contains the coefficients for the corresponding dependent variable.
    The second output is a vector of the root mean square of the residuals of
    each dependent variable. Both outputs are null if the system is
    underdetermined, e.g., the number of inputs is less than or equal to numX.

    Args:
      numX: The number of input dimensions.
      numY: The number of output dimensions.
    """

    return apifunction.ApiFunction.call_('Reducer.linearRegression', numX, numY)

  @staticmethod
  def max(
      numInputs: Optional[_IntegerType] = None,  # pylint: disable=invalid-name
  ) -> Reducer:
    """Returns a reducer that outputs the maximum value of its (first) input.

    If numInputs is greater than one, also outputs the corresponding values of
    the additional inputs.

    Args:
      numInputs: The number of inputs.
    """

    return apifunction.ApiFunction.call_('Reducer.max', numInputs)

  @staticmethod
  def mean() -> Reducer:
    """Returns a Reducer that computes the weighted arithmetic mean.

    Where applicable, the output name is "mean".
    """

    return apifunction.ApiFunction.call_('Reducer.mean')

  @staticmethod
  def median(
      # pylint: disable=invalid-name
      maxBuckets: Optional[_IntegerType] = None,
      minBucketWidth: Optional[_NumberType] = None,
      maxRaw: Optional[_IntegerType] = None,
      # pylint: enable=invalid-name
  ) -> Reducer:
    """Returns a reducer that will compute the median of the inputs.

    For small numbers of inputs (up to maxRaw) the median will be computed
    directly; for larger numbers of inputs the median will be derived from a
    histogram.

    Args:
      maxBuckets: The maximum number of buckets to use when building a
        histogram; will be rounded up to a power of 2.
      minBucketWidth: The minimum histogram bucket width, or null to allow any
        power of 2.
      maxRaw: The number of values to accumulate before building the initial
        histogram.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.median', maxBuckets, minBucketWidth, maxRaw
    )

  @staticmethod
  def min(
      numInputs: Optional[_IntegerType] = None,  # pylint: disable=invalid-name
  ) -> Reducer:
    """Returns a reducer that outputs the minimum value of its first input.

    If numInputs is greater than one, also outputs the corresponding values of
    the additional inputs.

    Args:
      numInputs: The number of inputs.
    """

    return apifunction.ApiFunction.call_('Reducer.min', numInputs)

  @staticmethod
  def minMax() -> Reducer:
    """Returns a Reducer that computes the minimum and maximum of its inputs.

    Where applicable, the outputs are named: "min", "max".
    """

    return apifunction.ApiFunction.call_('Reducer.minMax')

  @staticmethod
  def mode(
      # pylint: disable=invalid-name
      maxBuckets: Optional[_IntegerType] = None,
      minBucketWidth: Optional[_NumberType] = None,
      maxRaw: Optional[_IntegerType] = None,
      # pylint: enable=invalid-name
  ) -> Reducer:
    """Returns a reducer that will compute the mode of the inputs.

    For small numbers of inputs (up to maxRaw) the mode will be computed
    directly; for larger numbers of inputs the mode will be derived from a
    histogram.

    Args:
      maxBuckets: The maximum number of buckets to use when building a
        histogram; will be rounded up to a power of 2.
      minBucketWidth: The minimum histogram bucket width, or null to allow any
        power of 2.
      maxRaw: The number of values to accumulate before building the initial
        histogram.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.mode', maxBuckets, minBucketWidth, maxRaw
    )

  @staticmethod
  @deprecation.Deprecated('Use Reducer.anyNonZero().')
  def Or() -> Reducer:
    """Returns a Reducer that returns 1 if any inputs are not 0, 0 otherwise.

    Where applicable, the output name is "any".
    """

    return apifunction.ApiFunction.call_('Reducer.or')

  @staticmethod
  def pearsonsCorrelation() -> Reducer:
    """Returns a Pearson's correlation reducer.

    Creates a two-input reducer that computes Pearson's product-moment
    correlation coefficient and the 2-sided p-value test for correlation = 0.
    """

    return apifunction.ApiFunction.call_('Reducer.pearsonsCorrelation')

  @staticmethod
  def percentile(
      percentiles: _ListType,
      # pylint: disable=invalid-name
      outputNames: Optional[_ListType] = None,
      maxBuckets: Optional[_IntegerType] = None,
      minBucketWidth: Optional[_NumberType] = None,
      maxRaw: Optional[_IntegerType] = None,
      # pylint: enable=invalid-name
  ) -> Reducer:
    """Returns a reducer that will compute the specified percentiles.

    For example,given [0, 50, 100] will produce outputs named 'p0', 'p50', and
    'p100' with the min, median, and max respectively. For small numbers of
    inputs (up to maxRaw) the percentiles will be computed directly; for larger
    numbers of inputs the percentiles will be derived from a histogram.

    Args:
      percentiles: A list of numbers between 0 and 100.
      outputNames: A list of names for the outputs, or null to get default
        names.
      maxBuckets: The maximum number of buckets to use when building a
        histogram; will be rounded up to a power of 2.
      minBucketWidth: The minimum histogram bucket width, or null to allow any
        power of 2.
      maxRaw: The number of values to accumulate before building the initial
        histogram.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.percentile',
        percentiles,
        outputNames,
        maxBuckets,
        minBucketWidth,
        maxRaw,
    )

  @staticmethod
  def product() -> Reducer:
    """Returns a Reducer that computes the product of its inputs.

    Where applicable, the output name is "product".
    """

    return apifunction.ApiFunction.call_('Reducer.product')

  def repeat(self, count: _IntegerType) -> Reducer:
    """Returns a reducer by combining the count copies of the reducer.

    Output names are the same as the given reducer, but each is a list of the
    corresponding output from each of the reducers.

    Args:
      count: The number of copies of the reducer to combine.
    """

    return apifunction.ApiFunction.call_(self.name() + '.repeat', self, count)

  @staticmethod
  def ridgeRegression(
      numX: _IntegerType,  # pylint: disable=invalid-name
      numY: Optional[_IntegerType] = None,  # pylint: disable=invalid-name
      lambda_: Optional[_NumberType] = None,
      **kwargs,
  ) -> Reducer:
    # pylint: disable=g-doc-args
    """Returns a ridge regression reducer.

    Creates a reducer that computes a ridge regression with numX independent
    variables (not including constant) followed by numY dependent variables.
    Ridge regression is a form of Tikhonov regularization which shrinks the
    regression coefficients by imposing a penalty on their size. With this
    implementation of ridge regression there NO NEED to include a constant value
    for bias.

    The first output is a coefficients array with dimensions (numX + 1, numY);
    each column contains the coefficients for the corresponding dependent
    variable plus the intercept for the dependent variable in the last column.
    Additional outputs are a vector of the root mean square of the residuals of
    each dependent variable and a vector of p-values for each dependent
    variable. Outputs are null if the system is underdetermined, e.g., the
    number of inputs is less than numX + 1.

    Args:
      numX: the number of independent variables being regressed.
      numY: the number of dependent variables.
      lambda: Regularization parameter.
    """
    # pylint: enable=g-doc-args
    if kwargs:
      if lambda_ is not None:
        raise ValueError('lambda_ cannot be set when providing kwargs.')
      if kwargs.keys() != {'lambda'}:
        raise ValueError(
            f'Unexpected arguments: {list(kwargs.keys())}. Expected: lambda.'
        )
      lambda_ = kwargs['lambda']

    return apifunction.ApiFunction.call_(
        'Reducer.ridgeRegression', numX, numY, lambda_
    )

  @staticmethod
  def robustLinearRegression(
      numX: _IntegerType,  # pylint: disable=invalid-name
      numY: Optional[_IntegerType] = None,  # pylint: disable=invalid-name
      beta: Optional[_NumberType] = None,
  ) -> Reducer:
    """Returns a robust linear regression reducer.

    Creates a reducer that computes a robust least squares regression with numX
    independent variables and numY dependent variables, using iteratively
    reweighted least squares with the Talwar cost function. A point is
    considered an outlier if the RMS of residuals is greater than beta.

    Each input tuple will have values for the independent variables followed by
    the dependent variables.

    The first output is a coefficients array with dimensions (numX, numY); each
    column contains the coefficients for the corresponding dependent variable.
    The second is a vector of the root mean square of the residuals of each
    dependent variable. Both outputs are null if the system is underdetermined,
    e.g., the number of inputs is less than numX.

    Args:
      numX: The number of input dimensions.
      numY: The number of output dimensions.
      beta: Residual error outlier margin. If null, a default value will be
        computed.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.robustLinearRegression', numX, numY, beta
    )

  @staticmethod
  def sampleStdDev() -> Reducer:
    """Returns a reducer that computes the sample standard deviation.

    Where applicable, the output name is "stdDev".
    """

    return apifunction.ApiFunction.call_('Reducer.sampleStdDev')

  @staticmethod
  def sampleVariance() -> Reducer:
    """Returns a reducer that computes the sample variance of its inputs.

    Where applicable, the output name is "variance".
    """

    return apifunction.ApiFunction.call_('Reducer.sampleVariance')

  @staticmethod
  def sensSlope() -> Reducer:
    """Returns a two-input reducer that computes the Sen's slope estimator.

    The inputs are expected to be x data followed by y data. It returns two
    double values; the estimated slope and the offset.
    """

    return apifunction.ApiFunction.call_('Reducer.sensSlope')

  def setOutputs(self, outputs: _ListType) -> Reducer:
    """Returns a reducer with outputs renamed and/or removed.

    Args:
      outputs: The new output names; any output whose name is null or empty will
        be dropped.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.setOutputs', self, outputs
    )

  @staticmethod
  def skew() -> Reducer:
    """Returns a reducer that Computes the skewness of its inputs.

    Where applicable, the output name is "skew".
    """

    return apifunction.ApiFunction.call_('Reducer.skew')

  @staticmethod
  def spearmansCorrelation() -> Reducer:
    """Returns a two-input Spearman's rank-moment correlation reducer.

    See
    https://commons.apache.org/proper/commons-math/javadocs/api-3.6/org/apache/commons/math3/stat/correlation/SpearmansCorrelation.html
    for details.
    """

    return apifunction.ApiFunction.call_('Reducer.spearmansCorrelation')

  def splitWeights(self) -> Reducer:
    """Returns a reducer that splits each weighted input into two unweighted inputs."""

    return apifunction.ApiFunction.call_(self.name() + '.splitWeights', self)

  @staticmethod
  def stdDev() -> Reducer:
    """Returns a Reducer that computes the standard deviation of its inputs.

    Where applicable, the output name is "stdDev".
    """

    return apifunction.ApiFunction.call_('Reducer.stdDev')

  @staticmethod
  def sum() -> Reducer:
    """Returns a Reducer that computes the (weighted) sum of its inputs.

    Where applicable, the output name is "sum".
    """

    return apifunction.ApiFunction.call_('Reducer.sum')

  @staticmethod
  def toCollection(
      propertyNames: _ListType,  # pylint: disable=invalid-name
      numOptional: Optional[_IntegerType] = None  # pylint: disable=invalid-name
  ) -> Reducer:
    """Returns a reducer that collects its inputs into a FeatureCollection.

    Args:
      propertyNames: The property names that will be defined on each output
        feature; determines the number of reducer inputs.
      numOptional: The last numOptional inputs will be considered optional; the
        other inputs must be non-null or the input tuple will be dropped.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.toCollection', propertyNames, numOptional
    )

  @staticmethod
  def toList(
      # pylint: disable=invalid-name
      tupleSize: Optional[_IntegerType] = None,
      numOptional: Optional[_IntegerType] = None,
      # pylint: enable=invalid-name
  ) -> Reducer:
    """Returns a reducer that collects its inputs into a list.

    Args:
      tupleSize: The size of each output tuple, or null for no grouping. Also
        determines the number of inputs (null tupleSize has 1 input).
      numOptional: The last numOptional inputs will be considered optional; the
        other inputs must be non-null or the input tuple will be dropped.
    """

    return apifunction.ApiFunction.call_(
        'Reducer.toList', tupleSize, numOptional
    )

  def unweighted(self) -> Reducer:
    """Returns a Reducer with no weighted inputs."""

    return apifunction.ApiFunction.call_(self.name() + '.unweighted', self)

  @staticmethod
  def variance() -> Reducer:
    """Returns a Reducer that computes the variance of its inputs.

    Where applicable, the output name is "variance".
    """

    return apifunction.ApiFunction.call_('Reducer.variance')



"""Convenience functions and code used by ee/__init__.py.

These functions are in general re-exported from the "ee" module and should be
referenced from there (e.g., "ee.profilePrinting").
"""

# Using lowercase function naming to match the JavaScript names.
# pylint: disable=g-bad-name

import contextlib
import json
import sys
from typing import Any, Dict, Iterator, Optional, TextIO, Union

from google.auth import crypt
from google.oauth2 import service_account

from ee import apifunction
from ee import computedobject
from ee import data
from ee import oauth


def ServiceAccountCredentials(
    email: str, key_file: Optional[str] = None, key_data: Optional[str] = None
) -> service_account.Credentials:
  """Configure OAuth2 credentials for a Google Service Account.

  Args:
    email: The email address of the account for which to configure credentials.
        Ignored if key_file or key_data represents a JSON service account key.
    key_file: The path to a file containing the private key associated with
        the service account. Both JSON and PEM files are supported.
    key_data: Raw key data to use, if key_file is not specified.

  Returns:
    An OAuth2 credentials object.
  """

  # Assume anything that doesn't end in '.pem' is a JSON key.
  if key_file and not key_file.endswith('.pem'):
    return service_account.Credentials.from_service_account_file(
        key_file, scopes=oauth.SCOPES)

  # If 'key_data' can be decoded as JSON, it's probably a raw JSON key.
  if key_data:
    try:
      key_data = json.loads(key_data)
      return service_account.Credentials.from_service_account_info(
          key_data, scopes=oauth.SCOPES)
    except ValueError:
      # It may actually be a raw PEM string, we'll try that below.
      pass

  # Probably a PEM key - just read the file into 'key_data'.
  if key_file:
    with open(key_file, 'r') as file_:
      key_data = file_.read()

  # Raw PEM key.
  signer = crypt.RSASigner.from_string(key_data)
  return service_account.Credentials(
      signer, email, oauth.TOKEN_URI, scopes=oauth.SCOPES)


def call(
    func: Union[str, apifunction.ApiFunction], *args, **kwargs
) -> computedobject.ComputedObject:
  """Invoke the given algorithm with the specified args.

  Args:
    func: The function to call. Either an ee.Function object or the name of
        an API function.
    *args: The positional arguments to pass to the function.
    **kwargs: The named arguments to pass to the function.

  Returns:
    A ComputedObject representing the called function. If the signature
    specifies a recognized return type, the returned value will be cast
    to that type.
  """
  if isinstance(func, str):
    func = apifunction.ApiFunction.lookup(func)
  return func.call(*args, **kwargs)


# pylint: disable-next=redefined-builtin
def apply(
    func: Union[str, apifunction.ApiFunction], named_args: Dict[str, Any]
) -> computedobject.ComputedObject:
  """Call a function with a dictionary of named arguments.

  Args:
    func: The function to call. Either an ee.Function object or the name of
        an API function.
    named_args: A dictionary of arguments to the function.

  Returns:
    A ComputedObject representing the called function. If the signature
    specifies a recognized return type, the returned value will be cast
    to that type.
  """
  if isinstance(func, str):
    func = apifunction.ApiFunction.lookup(func)
  return func.apply(named_args)


@contextlib.contextmanager
def profilePrinting(destination: TextIO = sys.stderr) -> Iterator[None]:
  # pylint: disable=g-doc-return-or-yield
  """Returns a context manager that prints a profile of enclosed API calls.

  The profile will be printed when the context ends, whether or not any error
  occurred within the context.

  # Simple example:
  with ee.profilePrinting():
     print ee.Number(1).add(1).getInfo()

  Args:
    destination: A file-like object to which the profile text is written.
        Defaults to sys.stderr.

  """
  # Profile.getProfiles is `hidden`, so call it explicitly.
  getProfiles = apifunction.ApiFunction.lookup('Profile.getProfiles')

  profile_ids = []
  try:
    with data.profiling(profile_ids.append):
      yield
  finally:
    profile_text = getProfiles.call(ids=profile_ids).getInfo()
    destination.write(profile_text)



"""Representation of an Earth Engine FeatureCollection."""

# Using lowercase function naming to match the JavaScript names.
# pylint: disable=g-bad-name

from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Type, Union

from ee import _utils
from ee import apifunction
from ee import classifier
from ee import clusterer
from ee import collection
from ee import computedobject
from ee import data
from ee import deprecation
from ee import ee_exception
from ee import ee_list
from ee import ee_number
from ee import ee_string
from ee import ee_types
from ee import errormargin
from ee import feature
from ee import geometry
from ee import image
from ee import reducer

_ClassifierType = Union['classifier.Classifier', computedobject.ComputedObject]
_ClustererType = Union['clusterer.Clusterer', computedobject.ComputedObject]
_ErrorMarginType = Union[
    float,
    'ee_number.Number',
    errormargin.ErrorMargin,
    computedobject.ComputedObject,
]
_IntegerType = Union[int, 'ee_number.Number', computedobject.ComputedObject]
_ListType = Union[List[Any], Tuple[Any, Any], computedobject.ComputedObject]
_NumberType = Union[float, 'ee_number.Number', computedobject.ComputedObject]
_ReducerType = Union[reducer.Reducer, computedobject.ComputedObject]
_StringType = Union[str, 'ee_string.String', computedobject.ComputedObject]


class FeatureCollection(collection.Collection):
  """A representation of a FeatureCollection."""

  _initialized = False

  # Tell pytype to not complain about dynamic attributes.
  _HAS_DYNAMIC_ATTRIBUTES = True

  @_utils.accept_opt_prefix('opt_column')
  @deprecation.WarnForDeprecatedAsset('args')
  def __init__(
      self,
      args: Optional[
          Union[
              Dict[str, Any],
              List[Any],
              str,
              feature.Feature,
              geometry.Geometry,
              computedobject.ComputedObject,
          ]
      ],
      column: Optional[Any] = None,
  ):
    """Constructs a collection features.

    Args:
      args: constructor argument.  One of: 1) A string - assumed to be the name
        of a collection. 2) A geometry. 3) A feature. 4) An array of features.
        5) A GeoJSON FeatureCollection. 6) A computed object - reinterpreted as
        a collection.
      column: The name of the geometry column to use. Only useful with the
        string constructor.

    Raises:
      EEException: if passed something other than the above.
    """
    self.initialize()

    # Wrap geometries with features.
    if isinstance(args, geometry.Geometry):
      args = feature.Feature(args)

    # Wrap single features in an array.
    if isinstance(args, feature.Feature):
      args = [args]

    if ee_types.isString(args):
      # An ID.
      actual_args = {'tableId': args}
      if column:
        actual_args['geometryColumn'] = column
      super().__init__(
          apifunction.ApiFunction.lookup('Collection.loadTable'), actual_args)
    elif isinstance(args, (list, tuple)):
      # A list of features.
      super().__init__(
          apifunction.ApiFunction.lookup('Collection'), {
              'features': [feature.Feature(i) for i in args]
          })
    elif isinstance(args, ee_list.List):
      # A computed list of features.
      super().__init__(
          apifunction.ApiFunction.lookup('Collection'), {
              'features': args
          })
    elif isinstance(args, dict) and args.get('type') == self.name():
      # A GeoJSON FeatureCollection
      super().__init__(
          apifunction.ApiFunction.lookup('Collection'),
          {'features': [feature.Feature(i) for i in args.get('features', [])]})
    elif isinstance(args, computedobject.ComputedObject):
      # A custom object to reinterpret as a FeatureCollection.
      super().__init__(args.func, args.args, args.varName)
    else:
      raise ee_exception.EEException(
          'Unrecognized argument type to convert to a FeatureCollection: %s' %
          args)

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      super().initialize()
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  def getMapId(
      self, vis_params: Optional[Dict[str, Any]] = None
  ) -> Dict[str, Any]:
    """Fetch and return a map id and token, suitable for use in a Map overlay.

    Args:
      vis_params: The visualization parameters. Currently only one parameter,
          'color', containing a hex RGB color string is allowed.

    Returns:
      A map ID dictionary as described in ee.data.getMapId, including an
      additional 'image' field containing Collection.draw image wrapping a
      FeatureCollection containing this feature.
    """
    painted = apifunction.ApiFunction.apply_('Collection.draw', {
        'collection': self,
        'color': (vis_params or {}).get('color', '000000')
    })
    return painted.getMapId({})

  def getDownloadURL(
      self,
      filetype: Optional[str] = None,
      selectors: Optional[Any] = None,
      filename: Optional[str] = None,
  ) -> str:
    """Gets a download URL.

    When the URL is accessed, the FeatureCollection is downloaded in one of
    several formats.

    Args:
      filetype: The format of download, one of: "csv", "json", "geojson", "kml",
          "kmz" ("json" outputs GeoJSON). If unspecified, defaults to "csv".
      selectors: Feature property names used to select the attributes to be
          downloaded. If unspecified, all properties are included.
      filename: Name of the file to be downloaded; extension is appended by
          default. If unspecified, defaults to "table".

    Returns:
      A URL to download this FeatureCollection.
    """
    request = {}
    request['table'] = self
    if filetype is not None:
      request['format'] = filetype.upper()
    if filename is not None:
      request['filename'] = filename
    if selectors is not None:
      if isinstance(selectors, (list, tuple)):
        selectors = ','.join(selectors)
      request['selectors'] = selectors
    return data.makeTableDownloadUrl(data.getTableDownloadId(request))

  # Deprecated spelling to match the JS library.
  getDownloadUrl = deprecation.Deprecated('Use getDownloadURL().')(
      getDownloadURL)

  # TODO(user): How to handle type annotations for
  #  feature_collection.select('a', 'b', 'c')?
  # pylint: disable-next=keyword-arg-before-vararg
  def select(
      self,
      propertySelectors: Any,
      newProperties: Optional[Any] = None,
      retainGeometry: Union[bool, str] = True,
      *args,
  ) -> FeatureCollection:
    """Select properties from each feature in a collection.

    Args:
      propertySelectors: An array of names or regexes specifying the properties
          to select.
      newProperties: An array of strings specifying the new names for the
          selected properties.  If supplied, the length must match the number
          of properties selected.
      retainGeometry: A boolean.  When false, the result will have no geometry.
      *args: Selector elements as varargs.

    Returns:
      The feature collection with selected properties.
    """
    if len(args) or ee_types.isString(propertySelectors):
      args = list(args)
      if not isinstance(retainGeometry, bool):
        args.insert(0, retainGeometry)
      if newProperties is not None:
        args.insert(0, newProperties)
      args.insert(0, propertySelectors)
      return self.map(lambda feat: feat.select(args, None, True))
    else:
      return self.map(
          # pylint: disable-next=g-long-lambda
          lambda feat: feat.select(
              propertySelectors, newProperties, retainGeometry
          )
      )

  @staticmethod
  def name() -> str:
    return 'FeatureCollection'

  @staticmethod
  def elementType() -> Type[feature.Feature]:
    return feature.Feature

  def classify(
      self,
      classifier: _ClassifierType,
      outputName: Optional[_StringType] = None,  # pylint: disable=invalid-name
  ) -> FeatureCollection:
    """Returns the result of classifying each feature in a collection.

    Args:
      classifier: The classifier to use.
      outputName: The name of the output property to be added. This argument is
        ignored if the classifier has more than one output.

    Returns:
      An ee.FeatureCollection.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.classify', self, classifier, outputName
    )

  def cluster(
      self,
      clusterer: _ClustererType,
      outputName: Optional[_StringType] = None,  # pylint: disable=invalid-name
  ) -> FeatureCollection:
    """Returns the results of clustering each feature in a collection.

    Clusters each feature in a collection, adding a new column to each feature
    containing the cluster number to which it has been assigned.

    Args:
      clusterer: The clusterer to use.
      outputName: The name of the output property to be added.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.cluster', self, clusterer, outputName
    )

  def inverseDistance(
      self,
      range: _NumberType,
      propertyName: _StringType,  # pylint: disable=invalid-name
      mean: _NumberType,
      stdDev: _NumberType,  # pylint: disable=invalid-name
      gamma: Optional[_NumberType] = None,
      reducer: Optional[_ReducerType] = None,
  ) -> image.Image:
    """Returns an inverse-distance weighted estimate of the value at each pixel.

    Args:
      range: Size of the interpolation window (in meters).
      propertyName: Name of the numeric property to be estimated.
      mean: Global expected mean.
      stdDev: Global standard deviation.
      gamma: Determines how quickly the estimates tend towards the global mean.
      reducer: Reducer used to collapse the 'propertyName' value of overlapping
        points into a single value.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.inverseDistance',
        self,
        range,
        propertyName,
        mean,
        stdDev,
        gamma,
        reducer,
    )

  def kriging(
      self,
      propertyName: _StringType,  # pylint: disable=invalid-name
      shape: _StringType,
      range: _NumberType,
      sill: _NumberType,
      nugget: _NumberType,
      maxDistance: Optional[_NumberType] = None,  # pylint: disable=invalid-name
      reducer: Optional[_ReducerType] = None,
  ) -> image.Image:
    """Returns the results of sampling a Kriging estimator at each pixel.

    Args:
      propertyName: Property to be estimated (must be numeric).
      shape: Semivariogram shape (one of {exponential, gaussian, spherical}).
      range: Semivariogram range, in meters.
      sill: Semivariogram sill.
      nugget: Semivariogram nugget.
      maxDistance: Radius which determines which features are included in each
        pixel's computation, in meters. Defaults to the semivariogram's range.
      reducer: Reducer used to collapse the 'propertyName' value of overlapping
        points into a single value.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.kriging',
        self,
        propertyName,
        shape,
        range,
        sill,
        nugget,
        maxDistance,
        reducer,
    )

  def makeArray(
      self, properties: _ListType, name: Optional[_StringType] = None
  ) -> FeatureCollection:
    """Returns a collection with a 1-D Array property for each feature.

    Add a 1-D Array to each feature in a collection by combining a list of
    properties for each feature into a 1-D Array.

    All of the properties must be numeric values. If a feature doesn't contain
    all of the named properties, or any of them aren't numeric, the feature will
    be dropped from the resulting collection.

    Args:
      properties: The properties to select.
      name: The name of the new array property.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.makeArray', self, properties, name
    )

  @staticmethod
  def randomPoints(
      region: geometry.Geometry,
      points: Optional[_IntegerType] = None,
      seed: Optional[_IntegerType] = None,
      # pylint: disable-next=invalid-name
      maxError: Optional[_ErrorMarginType] = None,
  ) -> FeatureCollection:
    """Returns a collection of random points.

    Generates points that are uniformly random in the given geometry. If the
    geometry is two-dimensional (polygon or multi-polygon) then the returned
    points are uniformly distributed on the given region of the sphere. If the
    geometry is one-dimensional (linestrings), the returned points are
    interpolated uniformly along the geometry's edges. If the geometry has
    dimension zero (points), the returned points are sampled uniformly from the
    input points. If a multi-geometry of mixed dimension is given, points are
    sampled from the component geometries with the highest dimension.

    Args:
      region: The region to generate points for.
      points: The number of points to generate.
      seed: A seed for the random number generator.
      maxError: The maximum amount of error tolerated when performing any
        necessary reprojection.
    """

    return apifunction.ApiFunction.call_(
        'FeatureCollection.randomPoints', region, points, seed, maxError
    )



#!/usr/bin/env python3
"""Tests for the ee.Projection module."""

import json
from typing import Any, Dict

import unittest
import ee
from ee import apitestcase

CRS = 'crs'
TRANSFORM = 'transform'
TRANSFORM_WKT = 'transformWkt'

PROJECTION = 'Projection'

EPSG_4326 = 'EPSG:4326'
PROJECTION_EPSG_4326 = {
    'functionInvocationValue': {
        'functionName': 'Projection',
        'arguments': {'crs': {'constantValue': EPSG_4326}},
    }
}


def make_expression_graph(
    function_invocation_value: Dict[str, Any],
) -> Dict[str, Any]:
  return {
      'result': '0',
      'values': {'0': {'functionInvocationValue': function_invocation_value}},
  }


class ProjectionTest(apitestcase.ApiTestCase):

  def test_init(self):
    crs = EPSG_4326
    transform = [1, 2, 3, 4, 5, 6]

    projection = ee.Projection(crs, transform=transform)
    self.assertEqual({'value': 'fakeValue'}, projection.getInfo())

    projection_func = ee.ApiFunction.lookup('Projection')
    self.assertEqual(projection_func, projection.func)
    self.assertFalse(projection.isVariable())
    self.assertEqual(set(['crs', 'transform']), set(projection.args))
    self.assertEqual(EPSG_4326, projection.args['crs'])
    expected_transform = {
        'result': '0',
        'values': {'0': {'constantValue': [1, 2, 3, 4, 5, 6]}},
    }
    self.assertEqual(
        expected_transform,
        json.loads(projection.args['transform'].serialize()),
    )

  def test_init_transform_wkt(self):
    crs = EPSG_4326
    transform_wkt = (
        'PARAM_MT["Affine",'
        '  PARAMETER["num_row",3],'
        '  PARAMETER["num_col",3],'
        '  PARAMETER["elt_0_1",1],'
        '  PARAMETER["elt_0_2",2],'
        '  PARAMETER["elt_1_2",3]]'
    )

    projection = ee.Projection(crs, transformWkt=transform_wkt)
    self.assertEqual({'value': 'fakeValue'}, projection.getInfo())

    projection_func = ee.ApiFunction.lookup('Projection')
    self.assertEqual(projection_func, projection.func)
    self.assertEqual(set(['crs', 'transformWkt']), set(projection.args))
    self.assertEqual(EPSG_4326, projection.args['crs'])
    expected_transform_wkt = {
        'result': '0',
        'values': {
            '0': {
                'constantValue': (
                    'PARAM_MT["Affine",  PARAMETER["num_row",3], '
                    ' PARAMETER["num_col",3],  PARAMETER["elt_0_1",1], '
                    ' PARAMETER["elt_0_2",2],  PARAMETER["elt_1_2",3]]'
                )
            }
        },
    }
    self.assertEqual(
        expected_transform_wkt,
        json.loads(projection.args['transformWkt'].serialize()),
    )
    projection_func = ee.ApiFunction.lookup('Projection')
    self.assertEqual(projection_func, projection.func)

  def test_serialize(self):
    crs = 'EPSG:32610'
    transform = [1, 2, 3, 4, 5, 6]

    result = json.loads(ee.Projection(crs, transform=transform).serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'arguments': {
                        'crs': {'constantValue': crs},
                        'transform': {'constantValue': transform},
                    },
                    'functionName': PROJECTION,
                }
            }
        },
    }
    self.assertEqual(expect, result)

  def test_cast(self):
    crs = 'EPSG:32610'

    projection = ee.Projection(ee.Projection(crs))
    result = json.loads(projection.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'arguments': {
                        'crs': {'constantValue': crs},
                    },
                    'functionName': PROJECTION,
                }
            }
        },
    }
    self.assertEqual(expect, result)

  def test_at_scale(self):
    expect = make_expression_graph({
        'arguments': {
            'projection': PROJECTION_EPSG_4326,
            'meters': {'constantValue': 1},
        },
        'functionName': 'Projection.atScale',
    })
    expression = ee.Projection(EPSG_4326).atScale(1)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Projection(EPSG_4326).atScale(meters=1)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_crs(self):
    expect = make_expression_graph({
        'arguments': {
            'projection': PROJECTION_EPSG_4326,
        },
        'functionName': 'Projection.crs',
    })
    expression = ee.Projection(EPSG_4326).crs()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_nominal_scale(self):
    expect = make_expression_graph({
        'arguments': {
            'proj': PROJECTION_EPSG_4326,
        },
        'functionName': 'Projection.nominalScale',
    })
    expression = ee.Projection(EPSG_4326).nominalScale()
    result = json.loads(expression.serialize())
    print(result)
    self.assertEqual(expect, result)

  def test_scale(self):
    expect = make_expression_graph({
        'arguments': {
            'projection': PROJECTION_EPSG_4326,
            'x': {'constantValue': 1},
            'y': {'constantValue': 2},
        },
        'functionName': 'Projection.scale',
    })
    expression = ee.Projection(EPSG_4326).scale(1, 2)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Projection(EPSG_4326).scale(x=1, y=2)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_transform(self):
    expect = make_expression_graph({
        'arguments': {
            'projection': PROJECTION_EPSG_4326,
        },
        'functionName': 'Projection.transform',
    })
    expression = ee.Projection(EPSG_4326).transform()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_translate(self):
    expect = make_expression_graph({
        'arguments': {
            'projection': PROJECTION_EPSG_4326,
            'x': {'constantValue': 1},
            'y': {'constantValue': 2},
        },
        'functionName': 'Projection.translate',
    })
    expression = ee.Projection(EPSG_4326).translate(1, 2)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Projection(EPSG_4326).translate(x=1, y=2)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_wkt(self):
    expect = make_expression_graph({
        'arguments': {
            'projection': PROJECTION_EPSG_4326,
        },
        'functionName': 'Projection.wkt',
    })
    expression = ee.Projection(EPSG_4326).wkt()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Tests for the ee.PixelType module."""

import enum
import json
from typing import Any, Dict

import unittest
import ee
from ee import apitestcase


class Type(str, enum.Enum):
  DOUBLE = 'double'
  FLOAT = 'float'
  INT = 'int'


DIMENSIONS_KEY = 'dimensions'
MAX_VALUE_KEY = 'maxValue'
MIN_VALUE_KEY = 'minValue'
PRECISION_KEY = 'precision'

PIXELTYPE = 'PixelType'


def make_expression_graph(
    function_invocation_value: Dict[str, Any]
) -> Dict[str, Any]:
  return {
      'result': '0',
      'values': {'0': {'functionInvocationValue': function_invocation_value}},
  }


def pixeltype_function_expr(value: Type) -> Dict[str, Any]:
  return {
      'functionInvocationValue': {
          'functionName': 'PixelType',
          'arguments': {'precision': {'constantValue': value}},
      }
  }


def pixeltype_noargs_expr(type_name: str) -> Dict[str, Any]:
  return {
      'result': '0',
      'values': {
          '0': {
              'functionInvocationValue': {
                  'functionName': 'PixelType',
                  'arguments': {
                      'precision': {
                          'functionInvocationValue': {
                              'functionName': 'PixelType.' + type_name,
                              'arguments': {},
                          }
                      }
                  },
              }
          }
      },
  }


class PixelTypeTest(apitestcase.ApiTestCase):

  def test_int(self):
    precision = Type.INT
    min_value = 0
    max_value = 1
    dimensions = 2
    pixeltype = ee.PixelType(precision, min_value, max_value, dimensions)
    self.assertEqual({'value': 'fakeValue'}, pixeltype.getInfo())

    pixeltype_func = ee.ApiFunction.lookup('PixelType')
    self.assertEqual(pixeltype_func, pixeltype.func)

    self.assertFalse(pixeltype.isVariable())
    self.assertEqual(
        set([DIMENSIONS_KEY, MAX_VALUE_KEY, MIN_VALUE_KEY, PRECISION_KEY]),
        set(pixeltype.args),
    )
    expected_dimensions = {'result': '0', 'values': {'0': {'constantValue': 2}}}
    self.assertEqual(
        expected_dimensions,
        json.loads(pixeltype.args[DIMENSIONS_KEY].serialize()),
    )
    expected_max_value = {'result': '0', 'values': {'0': {'constantValue': 1}}}
    self.assertEqual(
        expected_max_value,
        json.loads(pixeltype.args[MAX_VALUE_KEY].serialize()),
    )
    expected_min_value = {'result': '0', 'values': {'0': {'constantValue': 0}}}
    self.assertEqual(
        expected_min_value,
        json.loads(pixeltype.args[MIN_VALUE_KEY].serialize()),
    )
    self.assertEqual(Type.INT, pixeltype.args[PRECISION_KEY])

    result = json.loads(pixeltype.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'arguments': {
                        DIMENSIONS_KEY: {'constantValue': dimensions},
                        MAX_VALUE_KEY: {'constantValue': max_value},
                        MIN_VALUE_KEY: {'constantValue': min_value},
                        PRECISION_KEY: {'constantValue': precision},
                    },
                    'functionName': PIXELTYPE,
                }
            }
        },
    }
    self.assertEqual(expect, result)

  def test_minimal_double(self):
    precision = Type.DOUBLE
    pixeltype = ee.PixelType(precision)
    result = json.loads(pixeltype.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'arguments': {PRECISION_KEY: {'constantValue': precision}},
                    'functionName': PIXELTYPE,
                }
            }
        },
    }
    self.assertEqual(expect, result)

  def test_float_named_args(self):
    precision = Type.FLOAT
    min_value = 3
    max_value = 4
    dimensions = 5
    pixeltype = ee.PixelType(
        dimensions=dimensions,
        precision=precision,
        minValue=min_value,
        maxValue=max_value,
    )
    result = json.loads(pixeltype.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'arguments': {
                        DIMENSIONS_KEY: {'constantValue': dimensions},
                        MAX_VALUE_KEY: {'constantValue': max_value},
                        MIN_VALUE_KEY: {'constantValue': min_value},
                        PRECISION_KEY: {'constantValue': precision},
                    },
                    'functionName': PIXELTYPE,
                }
            }
        },
    }
    self.assertEqual(expect, result)

    cast_result = json.loads(ee.PixelType(pixeltype).serialize())
    self.assertEqual(expect, cast_result)

  def test_float_computed_object_args(self):
    precision = Type.FLOAT
    min_value = 3
    max_value = 4
    dimensions = 5
    pixeltype = ee.PixelType(
        precision=ee.String(precision).cat(''),
        minValue=ee.Number(min_value),
        maxValue=ee.Number(max_value),
        dimensions=ee.Number(dimensions),
    )
    result = json.loads(pixeltype.serialize())
    expected_precision = {
        'functionInvocationValue': {
            'functionName': 'String.cat',
            'arguments': {
                'string1': {'constantValue': 'float'},
                'string2': {'constantValue': ''},
            },
        }
    }
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'arguments': {
                        DIMENSIONS_KEY: {'constantValue': dimensions},
                        MAX_VALUE_KEY: {'constantValue': max_value},
                        MIN_VALUE_KEY: {'constantValue': min_value},
                        PRECISION_KEY: expected_precision,
                    },
                    'functionName': PIXELTYPE,
                }
            }
        },
    }
    self.assertEqual(expect, result)

    cast_result = json.loads(ee.PixelType(pixeltype).serialize())
    self.assertEqual(expect, cast_result)

  def test_float_no_dimensions(self):
    precision = Type.FLOAT
    min_value = 0.1
    max_value = 0.2
    result = json.loads(
        ee.PixelType(precision, min_value, max_value).serialize()
    )
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'arguments': {
                        MAX_VALUE_KEY: {'constantValue': 0.2},
                        MIN_VALUE_KEY: {'constantValue': 0.1},
                        PRECISION_KEY: {'constantValue': precision},
                    },
                    'functionName': PIXELTYPE,
                }
            }
        },
    }
    self.assertEqual(expect, result)

  def test_dimensions(self):
    expect = make_expression_graph({
        'arguments': {
            'pixelType': pixeltype_function_expr(Type.FLOAT),
        },
        'functionName': 'PixelType.dimensions',
    })
    expression = ee.PixelType(Type.FLOAT).dimensions()
    result = json.loads(expression.serialize())

    self.assertEqual(expect, result)

  def test_maxValue(self):
    expect = make_expression_graph({
        'arguments': {
            'pixelType': pixeltype_function_expr(Type.FLOAT),
        },
        'functionName': 'PixelType.maxValue',
    })
    expression = ee.PixelType(Type.FLOAT).maxValue()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_minValue(self):
    expect = make_expression_graph({
        'arguments': {
            'pixelType': pixeltype_function_expr(Type.FLOAT),
        },
        'functionName': 'PixelType.minValue',
    })
    expression = ee.PixelType(Type.FLOAT).minValue()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_precision(self):
    expect = make_expression_graph({
        'arguments': {
            'pixelType': pixeltype_function_expr(Type.FLOAT),
        },
        'functionName': 'PixelType.precision',
    })
    expression = ee.PixelType(Type.FLOAT).precision()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_double(self):
    expect = pixeltype_noargs_expr('double')
    expression = ee.PixelType.double()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_float(self):
    expect = pixeltype_noargs_expr('float')
    expression = ee.PixelType.float()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_int16(self):
    expect = pixeltype_noargs_expr('int16')
    expression = ee.PixelType.int16()
    result = json.loads(expression.serialize())
    print(json.dumps(result, indent=2))
    self.assertEqual(expect, result)

  def test_int32(self):
    expect = pixeltype_noargs_expr('int32')
    expression = ee.PixelType.int32()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_int64(self):
    expect = pixeltype_noargs_expr('int64')
    expression = ee.PixelType.int64()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_int8(self):
    expect = pixeltype_noargs_expr('int8')
    expression = ee.PixelType.int8()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_uint16(self):
    expect = pixeltype_noargs_expr('uint16')
    expression = ee.PixelType.uint16()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_uint32(self):
    expect = pixeltype_noargs_expr('uint32')
    expression = ee.PixelType.uint32()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_uint8(self):
    expect = pixeltype_noargs_expr('uint8')
    expression = ee.PixelType.uint8()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Tests for the ee.Classifier module."""

import json
import sys
from typing import Any, Dict
import unittest

import unittest
import ee
from ee import apitestcase


_CLASSIFIER_JSON = {
    'result': '0',
    'values': {
        '0': {
            'functionInvocationValue': {
                'functionName': 'Classifier.smileNaiveBayes',
                'arguments': {'lambda': {'constantValue': 0.01}},
            }
        }
    },
}

# Result of ee.Classifier.load('a')
_CLASSIFIER_LOAD_A = {
    'functionInvocationValue': {
        'functionName': 'Classifier.load',
        'arguments': {'id': {'constantValue': 'a'}},
    }
}


def make_expression_graph(
    function_invocation_value: Dict[str, Any],
) -> Dict[str, Any]:
  return {
      'result': '0',
      'values': {'0': {'functionInvocationValue': function_invocation_value}},
  }


class ClassifierTest(apitestcase.ApiTestCase):

  def test_naive_bayes_args(self):
    classifier = ee.Classifier.smileNaiveBayes(0.01)
    self.assertEqual({'value': 'fakeValue'}, classifier.getInfo())

    naive_bayes_func = ee.ApiFunction.lookup('Classifier.smileNaiveBayes')
    self.assertEqual(naive_bayes_func, classifier.func)
    self.assertFalse(classifier.isVariable())

    result = json.loads(classifier.serialize())
    self.assertEqual(_CLASSIFIER_JSON, result)

  def test_square_kwargs(self):
    kwargs = {'lambda': 0.01}
    classifier = ee.Classifier.smileNaiveBayes(**kwargs)
    result = json.loads(classifier.serialize())
    self.assertEqual(_CLASSIFIER_JSON, result)

  def test_cast(self):
    classifier = ee.Classifier(ee.Classifier.smileNaiveBayes(0.01))
    result = json.loads(classifier.serialize())
    self.assertEqual(_CLASSIFIER_JSON, result)

  @unittest.skipIf(sys.version_info < (3, 10), 'Unsupported in Python <= 3.9')
  def test_no_args(self):
    message = (
        r'Classifier\.__init__\(\) missing 1 required positional argument:'
        r' \'classifier\''
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Classifier()  # pytype:disable=missing-parameter

  def test_wrong_type(self):
    message = (
        r'Classifier can only be used as a cast to Classifier\. Found <class'
        r' \'int\'>'
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Classifier(1234)  # pytype:disable=wrong-arg-types

  def test_amnh_maxent(self):
    categorical_names = ['a', 'b']
    output_format = 'cloglog'
    auto_feature = False
    linear = True
    quadratic = False
    product = True
    threshold = False
    hinge = True
    hinge_threshold = 1
    l2lq_threshold = 2
    lq2lqpt_threshold = 3
    add_samples_to_background = False
    add_all_samples_to_background = True
    beta_multiplier = False
    beta_hinge = 4.0
    beta_lqp = 5.0
    beta_categorical = 6.0
    beta_threshold = 7.0
    extrapolate = True
    do_clamp = False
    write_clamp_grid = True
    random_test_points = 8
    seed = 9
    expect = make_expression_graph({
        'arguments': {
            'categoricalNames': {'constantValue': categorical_names},
            'outputFormat': {'constantValue': output_format},
            'autoFeature': {'constantValue': auto_feature},
            'linear': {'constantValue': linear},
            'quadratic': {'constantValue': quadratic},
            'product': {'constantValue': product},
            'threshold': {'constantValue': threshold},
            'hinge': {'constantValue': hinge},
            'hingeThreshold': {'constantValue': hinge_threshold},
            'l2lqThreshold': {'constantValue': l2lq_threshold},
            'lq2lqptThreshold': {'constantValue': lq2lqpt_threshold},
            'addSamplesToBackground': {
                'constantValue': add_samples_to_background
            },
            'addAllSamplesToBackground': {
                'constantValue': add_all_samples_to_background
            },
            'betaMultiplier': {'constantValue': beta_multiplier},
            'betaHinge': {'constantValue': beta_hinge},
            'betaLqp': {'constantValue': beta_lqp},
            'betaCategorical': {'constantValue': beta_categorical},
            'betaThreshold': {'constantValue': beta_threshold},
            'extrapolate': {'constantValue': extrapolate},
            'doClamp': {'constantValue': do_clamp},
            'writeClampGrid': {'constantValue': write_clamp_grid},
            'randomTestPoints': {'constantValue': random_test_points},
            'seed': {'constantValue': seed},
        },
        'functionName': 'Classifier.amnhMaxent',
    })
    expression = ee.Classifier.amnhMaxent(
        categorical_names,
        output_format,
        auto_feature,
        linear,
        quadratic,
        product,
        threshold,
        hinge,
        hinge_threshold,
        l2lq_threshold,
        lq2lqpt_threshold,
        add_samples_to_background,
        add_all_samples_to_background,
        beta_multiplier,
        beta_hinge,
        beta_lqp,
        beta_categorical,
        beta_threshold,
        extrapolate,
        do_clamp,
        write_clamp_grid,
        random_test_points,
        seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.amnhMaxent(
        categoricalNames=categorical_names,
        outputFormat=output_format,
        autoFeature=auto_feature,
        linear=linear,
        quadratic=quadratic,
        product=product,
        threshold=threshold,
        hinge=hinge,
        hingeThreshold=hinge_threshold,
        l2lqThreshold=l2lq_threshold,
        lq2lqptThreshold=lq2lqpt_threshold,
        addSamplesToBackground=add_samples_to_background,
        addAllSamplesToBackground=add_all_samples_to_background,
        betaMultiplier=beta_multiplier,
        betaHinge=beta_hinge,
        betaLqp=beta_lqp,
        betaCategorical=beta_categorical,
        betaThreshold=beta_threshold,
        extrapolate=extrapolate,
        doClamp=do_clamp,
        writeClampGrid=write_clamp_grid,
        randomTestPoints=random_test_points,
        seed=seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_confusion_matrix(self):
    classifier = ee.Classifier.load('a')
    expect = make_expression_graph({
        'arguments': {'classifier': _CLASSIFIER_LOAD_A},
        'functionName': 'Classifier.confusionMatrix',
    })
    expression = classifier.confusionMatrix()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_decision_tree(self):
    tree_string = 'a'
    expect = make_expression_graph({
        'arguments': {
            'treeString': {'constantValue': tree_string},
        },
        'functionName': 'Classifier.decisionTree',
    })
    expression = ee.Classifier.decisionTree(tree_string)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.decisionTree(treeString=tree_string)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_decision_tree_ensemble(self):
    tree_strings = ['a', 'b']
    expect = make_expression_graph({
        'arguments': {
            'treeStrings': {'constantValue': tree_strings},
        },
        'functionName': 'Classifier.decisionTreeEnsemble',
    })
    expression = ee.Classifier.decisionTreeEnsemble(tree_strings)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.decisionTreeEnsemble(treeStrings=tree_strings)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_explain(self):
    classifier = ee.Classifier.load('a')
    expect = make_expression_graph({
        'arguments': {'classifier': _CLASSIFIER_LOAD_A},
        'functionName': 'Classifier.explain',
    })
    expression = classifier.explain()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_libsvm(self):
    decision_procedure = 'Voting'
    svm_type = 'NU_SVC'
    kernel_type = 'POLY'
    shrinking = True
    degree = 1
    gamma = 2.0
    coef0 = 3.0
    cost = 4.0
    nu = 5.0
    termination_epsilon = 6.0
    loss_epsilon = 7.0
    one_class = 0
    expect = make_expression_graph({
        'arguments': {
            'decisionProcedure': {'constantValue': decision_procedure},
            'svmType': {'constantValue': svm_type},
            'kernelType': {'constantValue': kernel_type},
            'shrinking': {'constantValue': shrinking},
            'degree': {'constantValue': degree},
            'gamma': {'constantValue': gamma},
            'coef0': {'constantValue': coef0},
            'cost': {'constantValue': cost},
            'nu': {'constantValue': nu},
            'terminationEpsilon': {'constantValue': termination_epsilon},
            'lossEpsilon': {'constantValue': loss_epsilon},
            'oneClass': {'constantValue': one_class},
        },
        'functionName': 'Classifier.libsvm',
    })
    expression = ee.Classifier.libsvm(
        decision_procedure,
        svm_type,
        kernel_type,
        shrinking,
        degree,
        gamma,
        coef0,
        cost,
        nu,
        termination_epsilon,
        loss_epsilon,
        one_class,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.libsvm(
        decisionProcedure=decision_procedure,
        svmType=svm_type,
        kernelType=kernel_type,
        shrinking=shrinking,
        degree=degree,
        gamma=gamma,
        coef0=coef0,
        cost=cost,
        nu=nu,
        terminationEpsilon=termination_epsilon,
        lossEpsilon=loss_epsilon,
        oneClass=one_class,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_load(self):
    id = 'a'
    expect = make_expression_graph({
        'arguments': {
            'id': {'constantValue': id},
        },
        'functionName': 'Classifier.load',
    })
    expression = ee.Classifier.load(id)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.load(id=id)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_minimum_distance(self):
    metric = 'cosine'
    k_nearest = 1
    expect = make_expression_graph({
        'arguments': {
            'metric': {'constantValue': metric},
            'kNearest': {'constantValue': k_nearest},
        },
        'functionName': 'Classifier.minimumDistance',
    })
    expression = ee.Classifier.minimumDistance(metric, k_nearest)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.minimumDistance(
        metric=metric, kNearest=k_nearest
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_mode(self):
    classifier = ee.Classifier.load('a')
    expect = make_expression_graph({
        'arguments': {'classifier': _CLASSIFIER_LOAD_A},
        'functionName': 'Classifier.mode',
    })
    expression = classifier.mode()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_schema(self):
    classifier = ee.Classifier.load('a')
    expect = make_expression_graph({
        'arguments': {'classifier': _CLASSIFIER_LOAD_A},
        'functionName': 'Classifier.schema',
    })
    expression = classifier.schema()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_set_output_mode(self):
    classifier = ee.Classifier.load('a')
    mode = 'PROBABILITY'
    expect = make_expression_graph({
        'arguments': {
            'classifier': _CLASSIFIER_LOAD_A,
            'mode': {'constantValue': mode},
        },
        'functionName': 'Classifier.setOutputMode',
    })
    expression = classifier.setOutputMode(mode)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = classifier.setOutputMode(mode=mode)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_smile_cart(self):
    max_nodes = 1
    min_leaf_population = 2
    expect = make_expression_graph({
        'arguments': {
            'maxNodes': {'constantValue': max_nodes},
            'minLeafPopulation': {'constantValue': min_leaf_population},
        },
        'functionName': 'Classifier.smileCart',
    })
    expression = ee.Classifier.smileCart(max_nodes, min_leaf_population)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.smileCart(
        maxNodes=max_nodes, minLeafPopulation=min_leaf_population
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_smile_gradient_tree_boost(self):
    number_of_trees = 1
    shrinkage = 0.2
    sampling_rate = 0.3
    max_nodes = 4
    loss = 'Huber'
    seed = 5
    expect = make_expression_graph({
        'arguments': {
            'numberOfTrees': {'constantValue': number_of_trees},
            'shrinkage': {'constantValue': shrinkage},
            'samplingRate': {'constantValue': sampling_rate},
            'maxNodes': {'constantValue': max_nodes},
            'loss': {'constantValue': loss},
            'seed': {'constantValue': seed},
        },
        'functionName': 'Classifier.smileGradientTreeBoost',
    })
    expression = ee.Classifier.smileGradientTreeBoost(
        number_of_trees, shrinkage, sampling_rate, max_nodes, loss, seed
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.smileGradientTreeBoost(
        numberOfTrees=number_of_trees,
        shrinkage=shrinkage,
        samplingRate=sampling_rate,
        maxNodes=max_nodes,
        loss=loss,
        seed=seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_smile_knn(self):
    k = 1
    search_method = 'AUTO'
    metric = 'EUCLIDEAN'
    expect = make_expression_graph({
        'arguments': {
            'k': {'constantValue': k},
            'searchMethod': {'constantValue': search_method},
            'metric': {'constantValue': metric},
        },
        'functionName': 'Classifier.smileKNN',
    })
    expression = ee.Classifier.smileKNN(k, search_method, metric)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.smileKNN(
        k=k, searchMethod=search_method, metric=metric
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_smile_naive_bayes(self):
    lambda_ = 1
    expect = make_expression_graph({
        'arguments': {
            'lambda': {'constantValue': lambda_},
        },
        'functionName': 'Classifier.smileNaiveBayes',
    })
    expression = ee.Classifier.smileNaiveBayes(lambda_)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    # 'lambda' is a reserved word in Python.
    args = {'lambda': lambda_}
    expression = ee.Classifier.smileNaiveBayes(**args)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_smile_naive_bayes_with_bad_kwargs(self):
    message = r"Unexpected arguments: \['bad_arg'\]\. Expected: lambda."
    with self.assertRaisesRegex(ValueError, message):
      ee.Classifier.smileNaiveBayes(bad_arg=1)

  def test_smile_random_forest(self):
    number_of_trees = 1
    variables_per_split = True
    min_leaf_population = 2
    bag_fraction = 0.5
    max_nodes = False
    seed = 3
    expect = make_expression_graph({
        'arguments': {
            'numberOfTrees': {'constantValue': number_of_trees},
            'variablesPerSplit': {'constantValue': variables_per_split},
            'minLeafPopulation': {'constantValue': min_leaf_population},
            'bagFraction': {'constantValue': bag_fraction},
            'maxNodes': {'constantValue': max_nodes},
            'seed': {'constantValue': seed},
        },
        'functionName': 'Classifier.smileRandomForest',
    })
    expression = ee.Classifier.smileRandomForest(
        number_of_trees,
        variables_per_split,
        min_leaf_population,
        bag_fraction,
        max_nodes,
        seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.smileRandomForest(
        numberOfTrees=number_of_trees,
        variablesPerSplit=variables_per_split,
        minLeafPopulation=min_leaf_population,
        bagFraction=bag_fraction,
        maxNodes=max_nodes,
        seed=seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_spectral_region(self):
    coordinates = [[1, 2], [3, 4]]
    schema = ['a', 'b']
    expect = make_expression_graph({
        'arguments': {
            'coordinates': {'constantValue': coordinates},
            'schema': {'constantValue': schema},
        },
        'functionName': 'Classifier.spectralRegion',
    })
    expression = ee.Classifier.spectralRegion(coordinates, schema)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Classifier.spectralRegion(
        coordinates=coordinates, schema=schema
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_train(self):
    classifier = ee.Classifier.load('a')
    features = ee.FeatureCollection('b')
    class_property = 'c'
    input_properties = ['d', 'e']
    subsampling = 1
    subsampling_seed = 2
    expect = make_expression_graph({
        'arguments': {
            'classifier': _CLASSIFIER_LOAD_A,
            'features': {
                'functionInvocationValue': {
                    'functionName': 'Collection.loadTable',
                    'arguments': {'tableId': {'constantValue': 'b'}},
                }
            },
            'classProperty': {'constantValue': class_property},
            'inputProperties': {'constantValue': input_properties},
            'subsampling': {'constantValue': subsampling},
            'subsamplingSeed': {'constantValue': subsampling_seed},
        },
        'functionName': 'Classifier.train',
    })
    expression = classifier.train(
        features,
        class_property,
        input_properties,
        subsampling,
        subsampling_seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = classifier.train(
        features=features,
        classProperty=class_property,
        inputProperties=input_properties,
        subsampling=subsampling,
        subsamplingSeed=subsampling_seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)


if __name__ == '__main__':
  unittest.main()
